/******************************************************************************
 * MODULE       : sensors.cpp
 * FUNCTION     : Hardware Abstraction Layer of Sensor service in Android Framework
 *              : This is source file of base
 * AUTHOR       : Masafumi Seike <masafumi.seike@dsn.rohm.co.jp >
 * PROGRAMMED   : Sensor Application Development Group, ROHM CO.,LTD.
 * MODIFICATION : Modified by ROHM, JUN/08/2012
 * REMARKS      :
 * COPYRIGHT    : Copyright (C) 2008 The Android Open Source Project
 *              : Copyright (C) 2011 - 2012 ROHM CO.,LTD.
 *              : Licensed under the Apache License, Version 2.0 (the "License");
 *              : you may not use this file except in compliance with the License.
 *              : You may obtain a copy of the License at
 *              :
 *              :      http://www.apache.org/licenses/LICENSE-2.0
 *              :
 *              : Unless required by applicable law or agreed to in writing, software
 *              : distributed under the License is distributed on an "AS IS" BASIS,
 *              : WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *              : See the License for the specific language governing permissions and
 *              : limitations under the License.
 *****************************************************************************/
/*****************************************************************************
 * Constant definition                                                       *
 *****************************************************************************/
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))

/*****************************************************************************
 * Include                                                                   *
 *****************************************************************************/
#include <poll.h>
#include <cutils/log.h>
#include <fcntl.h>
#include <errno.h>
#include "sensors.h"
#include "SensorBase.h"
#include "AmbientLightSensor.h"

/*****************************************************************************
 * Prototype declaration for private functions                               *
 *****************************************************************************/
static int open_sensors(const struct hw_module_t *module, const char *id, struct hw_device_t **device);
static int sensors__get_sensors_list(struct sensors_module_t *module, struct sensor_t const **list);

/*****************************************************************************
 * Structure declaration for each module                                     *
 *****************************************************************************/
/* Hardware module */
static struct hw_module_methods_t sensors_module_methods = {
    open : open_sensors
};

/* The SENSORS Module */
struct sensors_module_t HAL_MODULE_INFO_SYM = {
    common : {
        tag           : HARDWARE_MODULE_TAG,
        version_major : MAJOR_VERSION(BH1780_HAL_VER),
        version_minor : MINOR_VERSION(BH1780_HAL_VER),
        id            : SENSORS_HARDWARE_MODULE_ID,
        name          : "ROHM SENSORS MODULE",
        author        : "ROHM SEMICONDUCTOR",
        methods       : &sensors_module_methods,
        dso           : 0,          // not use
        reserved      : {0, },      // not use
    },
    get_sensors_list : sensors__get_sensors_list,
};

/* All supported sensors list */

/** MODULE REGISTRATION SUPPORT
 **
 ** This is required so that hardware/libhardware/hardware.c
 ** will dlopen() this library appropriately.
 **/

/*
 * the following is the list of all supported sensors.
 * this table is used to build sSensorList declared below
 * according to which hardware sensors are reported as
 * available from the emulator (see get_sensors_list below)
 *
 */
static const struct sensor_t sSensorList[] = {
    {
        name       : "BH1780 Ambient Light Sensor",
        vendor     : "ROHM SEMICONDUCTOR",
        version    : 1,
        handle     : SENSORS_LIGHT_HANDLE,
        type       : SENSOR_TYPE_LIGHT,
        maxRange   : 65535,
        resolution : 1,
        power      : 0.2,
        minDelay   : 125,
        reserved   : {}
    },
};

/*****************************************************************************
 * structure                                                                 *
 *****************************************************************************/
/* poll structure of sensor */
struct sensors_poll_context_t {
    struct sensors_poll_device_t device; // must be first

    sensors_poll_context_t();
    ~sensors_poll_context_t();
    int activate(int handle, int enabled);
    int setDelay(int handle, int64_t ns);
    int pollEvents(sensors_event_t *data, int count);

private:
    enum {
        light = 0,
        numSensorDrivers,
        numFds,
    };

    static const size_t wake         = numFds - 1;
    static const char   WAKE_MESSAGE = 'W';
    struct pollfd       mPollFds[numFds];
    int                 mWritePipeFd;
    SensorBase          *mSensors[numSensorDrivers];

    int handleToDriver(int handle) const {
        switch (handle) {
        case SENSORS_LIGHT_HANDLE:
            return (light);
        default:
            break;
        }
        return (-EINVAL);
    }
};

/*****************************************************************************
 * Functions                                                                 *
 *****************************************************************************/
/******************************************************************************
 * NAME       : sensors__get_sensors_list
 * FUNCTION   : get all sensors list
 * REMARKS    :
 *****************************************************************************/
static int sensors__get_sensors_list(struct sensors_module_t *module, struct sensor_t const **list)
{
    *list = sSensorList;
    return (ARRAY_SIZE(sSensorList));
}

/******************************************************************************
 * NAME       : sensors_poll_context_t::sensors_poll_context_t
 * FUNCTION   : make sensor area and initialize
 * REMARKS    :
 *****************************************************************************/
sensors_poll_context_t::sensors_poll_context_t()
{
    mSensors[light] = NULL;
    mSensors[light] = new AmbientLightSensor();
    if (mSensors[light] == 0) {
        return;
    }
    mPollFds[light].fd      = mSensors[light]->getFd();
    mPollFds[light].events  = POLLIN;
    mPollFds[light].revents = 0;

    int wakeFds[2];
    int result = pipe(wakeFds);
    LOGE_IF(result < 0, "error creating wake pipe (%s)", strerror(errno));
    fcntl(wakeFds[0], F_SETFL, O_NONBLOCK);
    fcntl(wakeFds[1], F_SETFL, O_NONBLOCK);
    mWritePipeFd = wakeFds[1];

    mPollFds[wake].fd      = wakeFds[0];
    mPollFds[wake].events  = POLLIN;
    mPollFds[wake].revents = 0;

    return;
}

/******************************************************************************
 * NAME       : sensors_poll_context_t::~sensors_poll_context_t
 * FUNCTION   : free sensor area
 * REMARKS    :
 *****************************************************************************/
sensors_poll_context_t::~sensors_poll_context_t() {

    for (int i = 0; i < numSensorDrivers; i++) {
        if (mSensors[i] == NULL) {
            continue;
        }
        delete mSensors[i];
    }
    close(mPollFds[wake].fd);
    close(mWritePipeFd);

    return;
}

/******************************************************************************
 * NAME       : sensors_poll_context_t::activate
 * FUNCTION   : call enable function of each sensor
 * REMARKS    :
 *****************************************************************************/
int sensors_poll_context_t::activate(int handle, int enabled) {

    int index = handleToDriver(handle);
    if (index < 0) {
        return (index);
    }
    if (mSensors[index] == NULL) {
        return (-1);
    }
    int err = mSensors[index]->enable(handle, enabled);
    if (enabled && !err) {
        const char wakeMessage(WAKE_MESSAGE);
        int        result = write(mWritePipeFd, &wakeMessage, 1);
        LOGE_IF(result < 0, "error sending wake message (%s)", strerror(errno));
    }

    return (err);
}

/******************************************************************************
 * NAME       : sensors_poll_context_t::setDelay
 * FUNCTION   : call setDelay function of each sensor
 * REMARKS    :
 *****************************************************************************/
int sensors_poll_context_t::setDelay(int handle, int64_t ns) {

    int index = handleToDriver(handle);
    if (index < 0) {
        return (index);
    }
    if (mSensors[index] == NULL) {
        return (-1);
    }
    return (mSensors[index]->setDelay(handle, ns));
}

/******************************************************************************
 * NAME       : sensors_poll_context_t::pollEvents
 * FUNCTION   : get readEvents and data
 * REMARKS    :
 *****************************************************************************/
int sensors_poll_context_t::pollEvents(sensors_event_t *data, int count)
{
    int nbEvents = 0;
    int n        = 0;

    do {
        // see if we have some leftover from the last poll()
        for (int i = 0; count && i < numSensorDrivers; i++) {
            if (mSensors[i] == NULL) {
                continue;
            }
            SensorBase *const sensor(mSensors[i]);
            if ((mPollFds[i].revents & POLLIN) || (sensor->hasPendingEvents())) {
                int nb = sensor->readEvents(data, count);
                if (nb < count) {
                    // no more data for this sensor
                    mPollFds[i].revents = 0;
                }
                count    -= nb;
                nbEvents += nb;
                data     += nb;
            }
        }

        if (count) {
            // we still have some room, so try to see if we can get
            // some events immediately or just wait if we don't have
            // anything to return
            n = poll(mPollFds, numFds, nbEvents ? 0 : -1);
            if (n < 0) {
                LOGE("poll() failed (%s)", strerror(errno));
                return (-errno);
            }
            if (mPollFds[wake].revents & POLLIN) {
                char msg;
                int  result = read(mPollFds[wake].fd, &msg, 1);
                LOGE_IF(result < 0, "error reading from wake pipe (%s)", strerror(errno));
                LOGE_IF(msg != WAKE_MESSAGE, "unknown message on wake queue (0x%02x)", int(msg));
                mPollFds[wake].revents = 0;
            }
        }
        // if we have events and space, go read them
    } while (n && count);

    return (nbEvents);
}

/******************************************************************************
 * NAME       : poll__close
 * FUNCTION   : free memory.
 *            : be call from framework.
 * REMARKS    :
 *****************************************************************************/
static int poll__close(struct hw_device_t *dev)
{
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    if (ctx) {
        delete ctx;
    }
    return (0);
}

/******************************************************************************
 * NAME       : poll__activate
 * FUNCTION   : active device.
 *            : be call from framework.
 * REMARKS    :
 *****************************************************************************/
static int poll__activate(struct sensors_poll_device_t *dev, int handle, int enabled)
{
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return (ctx->activate(handle, enabled));
}

/******************************************************************************
 * NAME       : poll__setDelay
 * FUNCTION   : set delay time.
 *            : be call from framework.
 * REMARKS    :
 *****************************************************************************/
static int poll__setDelay(struct sensors_poll_device_t *dev, int handle, int64_t ns)
{
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return (ctx->setDelay(handle, ns));
}

/******************************************************************************
 * NAME       : poll__poll
 * FUNCTION   : call poll function.
 *            : be call from framework.
 * REMARKS    :
 *****************************************************************************/
static int poll__poll(struct sensors_poll_device_t *dev, sensors_event_t *data, int count)
{
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return (ctx->pollEvents(data, count));
}

/******************************************************************************
 * NAME       : open_sensors
 * FUNCTION   : Open a new instance of a sensor device using name
 *            : be call from framework.
 * REMARKS    :
 *****************************************************************************/
static int open_sensors(const struct hw_module_t *module, const char *id, struct hw_device_t **device)
{
    sensors_poll_context_t *dev = NULL;

    dev = new sensors_poll_context_t();
    if (dev == NULL) {
        return (-1);
    }

    memset(&dev->device, 0, sizeof(sensors_poll_device_t));
    dev->device.common.tag     = HARDWARE_DEVICE_TAG;
    dev->device.common.version = 0;
    dev->device.common.module  = const_cast<hw_module_t *>(module);
    dev->device.common.close   = poll__close;
    dev->device.activate       = poll__activate;
    dev->device.setDelay       = poll__setDelay;
    dev->device.poll           = poll__poll;

    *device = &dev->device.common;

    return (0);
}

